//**********************************************************************
// This source code is provided as an example.  No warranty is expressed
// or implied.  It is your responsibility to determine its suitability
// to your application, to modify it as necessary, and to validate
// its correctness.
//
// Licensed Materials - Property of IBM
// (c) Copyright IBM Corporation 2000. All Rights Reserved.
//
// US Government Users Restricted Rights - Use, duplication, or
// disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//**********************************************************************
// Revision / History
// Defect     Date       Who      Description
// ---------- 01/03/2000 ldbader  Module created
//**********************************************************************

package com.ibm.etill.ldbcardcassette;


import  java.sql.Connection;
import  java.sql.PreparedStatement;
import  java.sql.ResultSet;
import  java.sql.SQLException;
import  java.util.Enumeration;
import  java.util.GregorianCalendar;
import  java.util.Vector;

import  com.ibm.etill.framework.archive.ETillArchive;
import  com.ibm.etill.framework.cassette.APIResponse;
import  com.ibm.etill.framework.cassette.BatchOpenRequest;
import  com.ibm.etill.framework.cassette.BatchCloseRequest;
import  com.ibm.etill.framework.cassette.BatchPurgeRequest;
import  com.ibm.etill.framework.cassette.Cassette;
import  com.ibm.etill.framework.cassette.CassetteBatch;
import  com.ibm.etill.framework.cassette.CassetteRequest;
import  com.ibm.etill.framework.cassette.CassetteWorkItem;
import  com.ibm.etill.framework.cassette.DepositReversalRequest;
import  com.ibm.etill.framework.cassette.DeleteBatchRequest;
import  com.ibm.etill.framework.cassette.RefundReversalRequest;
import  com.ibm.etill.framework.log.ErrorLog;
import  com.ibm.etill.framework.log.ETillCassetteException;
import  com.ibm.etill.framework.log.Trace;
import  com.ibm.etill.framework.payapi.Batch;
import  com.ibm.etill.framework.payapi.BatchTotalsItem;
import  com.ibm.etill.framework.payapi.Credit;
import  com.ibm.etill.framework.payapi.ETillAbortOperation;
import  com.ibm.etill.framework.payapi.Order;
import  com.ibm.etill.framework.payapi.ParameterTable;
import  com.ibm.etill.framework.payapi.Payment;
import  com.ibm.etill.framework.payapi.TransactionKey;
import  com.ibm.etill.framework.supervisor.Supervisor;
import  com.ibm.etill.framework.supervisor.TimeableTransaction;


/***********************************************************************
* The cassette batch object is responsible for performing all the
* cassette's batch requests (BatchOpen, BatchClose) and to manage
* the state of the framework batch object.
***********************************************************************/
public final class LdbCardBatch
   implements CassetteBatch, LdbCardConstants {

   private static final String COPYRIGHT =
      "(c) Copyright IBM Corporation 2000 - All Rights Reserved.";

   //<gutter>
   //</gutter>

   //*******************************************************************
   // This is the cassette batch object's instance data.
   //******************************************************************/
   /********************************************************************
   * This cassette batch attribute contains the reference back to the
   * cassette that owns the batch.
   ********************************************************************/
   private LdbCardCassette   objCassette            = null;
   /********************************************************************
   * This cassette batch attribute contains the reference back to the
   * associated framework batch object.
   ********************************************************************/
   private Batch             objFrameworkBatch      = null;
   /********************************************************************
   * This cassette batch attribute contains the reference to the
   * cassette account object that will be used to perform transactions
   * for this batch.
   ********************************************************************/
   private LdbCardAccount    objCassetteAccount     = null;
   /********************************************************************
   * This cassette batch attribute contains the approval code.  The
   * approval code is supplied by the payment engine when a batch
   * transaction is successfully performed.
   ********************************************************************/
   private String            objApprovalCode        = null;
   /********************************************************************
   * This cassette batch attribute contains the decline reason. The
   * decline reason, originally small and therefore cryptic, was
   * designed to be displayed on a card swipe terminal at a point of
   * sale terminal.  In this cassette, it provides detailed information
   * about the cause of the failure.  (Since the text is provided by
   * the financial network, different for each network, and not
   * standardized in any way, it is impossible to translate these text
   * strings into return codes.)
   ********************************************************************/
   private String            objDeclineReason       = null;
   /********************************************************************
   * This cassette batch attribute contains an indicator used to
   * determine if the batch can be used for new transactions.  For
   * example, if the batch failed to close (for example, because the
   * batch did not balance) the batch is returned to the BATCH_OPEN
   * state but should not be used for future deposits or refunds.
   * (Adding new transactions to a batch that is out of balance will not
   * help and may harm the batch.)  This indicator is set to 1 (true)
   * whenever an automatic batch close is attempted or whenever the
   * merchant explicitly issues a BatchClose request for this batch.
   ********************************************************************/
   private Integer           objClosedToTrans       = new Integer(0);

   /********************************************************************
   * This cassette batch attribute is contains a reference to a
   * CassetteWorkItem that is used to schedule an automatic batch
   * close.  If this value is null, there is no automatic bach close
   * in progress.
   ********************************************************************/
   private AutoCloseWorkItem objAutoCloseWorkItem   = null;

   /********************************************************************
   * This cassette batch attribute is contains a reference to a
   * CassetteWorkItem that is used to schedule an automatic batch
   * close at a particular time of day.  If this value is null, no
   * time of day close has been scheduled.
   ********************************************************************/
   private TimeOfDayWorkItem objTimerWorkItem       = null;

   /********************************************************************
   * This constructor initializes a new cassette batch object.  A
   * reference to the cassette that owns this batch and a reference to
   * associated framework's batch object is set.
   *
   * @param     argCassette - The cassette that owns this batch.
   * @param     argFrameworkBatch - The framework's batch object.
   * @param     argProtocolData - The protocol data parameter values
   *            provided by the merchant system.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public LdbCardBatch(
      LdbCardCassette   argCassette,
      Batch             argFrameworkBatch,
      ParameterTable    argProtocolData
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.LdbCardBatch()"
         );
      }

      this.objCassette        = argCassette;
      this.objFrameworkBatch  = argFrameworkBatch;
      this.objCassetteAccount =
         LdbCardAccount.retrieveAccount(
            this.objFrameworkBatch.getMerchantName()
         );

      // Set attributes in the framework batch object.  In this case,
      // the attributes are set to inform the framework that this
      // cassette supports batch close and purge requests.
      this.objFrameworkBatch.setCloseAllowed(true);
      this.objFrameworkBatch.setPurgeAllowed(true);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.LdbCardBatch()"
         );
      }
   }

   /********************************************************************
   * This constructor initializes an existing cassette batch object.
   * A reference to the associated framework's batch object is set
   * and the remaining attributes are resurrected from the cassette's
   * batch database table.
   *
   * @param     argCassette - The cassette that owns this batch.
   * @param     argFrameworkBatch - The framework's batch object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public LdbCardBatch(
      LdbCardCassette   argCassette,
      Batch             argFrameworkBatch
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.LdbCardBatch()"
         );
      }

      this.objCassette        = argCassette;
      this.objFrameworkBatch  = argFrameworkBatch;
      this.objCassetteAccount =
         LdbCardAccount.retrieveAccount(
            this.objFrameworkBatch.getMerchantName()
         );
      this.resurrectBatch();


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.LdbCardBatch()"
         );
      }
   }

   /********************************************************************
   *  This method is used in two cases.
   * <ol>
   * <li> The cassette account object calls this method after creating a
   * new batch.
   * <li> The cassette object calls this method when it receives an
   * administration request with the START_CASSETTE token.
   * </ol>
   * <p>
   * In either case, this method will start the cassette batch's service
   * threads if necessary.
   * <p>
   * NOTE:  This method <b>must</b> be called whenever a cassette batch
   * object is constructed.  Unfortunately, it cannot be called by the
   * constructors.  The cassette batch object is constructed before
   * the framework batch object is completely initialized and the
   * framework batch object must be completely initialized before
   * this method is called.  In order to start a service thread,
   * a BatchCloseRequest must be constructed.  This will fail if the
   * read and write locks in the framework batch object have not
   * been initialized.
   ********************************************************************/
   public void initializeServiceThreadsIfNecessary() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            " LdbCardBatch.initializeServiceThreadsIfNecessary()"
         );
      }

      // Service threads are only started for open batches.  If we
      // are initializing a cassette batch object that was resurrected
      // in any other state, there is nothing to do.
      if (this.objFrameworkBatch.getCurrentState() == BATCH_OPEN) {
         // Check the batch state to see if the batch is closed to new
         // transactions due to an automatic batch close.  If so, an
         // automatic batch close was lost and must be rescheduled.
         if (this.isBatchClosedForAutomaticClose()) {
            // Given the value of this indicator, assume an automatic close
            // was sent to the service queue but was never processed (most
            // likely because the Payment Manager failed before it could
            // be removed from the service thread queue).  In this case,
            // the automatic close is rescheduled.
            this.scheduleAutoClose();
         } else {
            // Check to see if this batch is open for new transactions.
            // If the batch is closed to new transactions, it is because
            // a batch close failed and therefore a time of day close
            // should not be scheduled.
            if (!this.isBatchClosedToNewTransactions()) {
               // If the merchant has requested an automatic daily
               // close, schedule the work item that will perform the close
               // at the appropriate time.
               this.scheduleTimeOfDayCloseIfRequired();
            }
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            " LdbCardBatch.initializeServiceThreadsIfNecessary()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework to obtain the cassette object
   * for the cassette that owns this batch.
   *
   * @return    - The cassette object for the cassette that owns this
   * batch
   ********************************************************************/
   public Cassette cassette() {
      return this.objCassette;
   }

   /********************************************************************
    * This method is used by the cassette object to resurrect an
    * existing cassette batch from the cassette batch database table.
    * <p>
    * This method is <b>not</b> part of the Archivable interface.  The
    * methods for the Archivable interface are passed a JDBC database
    * connection for the operation.  In this case, an ETillArchive
    * object is constructed to provide a JDBC databaes connection.
    *
    * @exception ETillAbortOperation - any failure results in this
    * exception.
    *******************************************************************/
   private void resurrectBatch() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.resurrectBatch()"
         );
      }

      // The SQL select statement to obtain the cassette batch data.
      String varSelect =
         " SELECT * FROM " +
         ETillArchive.geteTillDatabaseOwner() + "." + TBL_BATCH +
         " WHERE " +
         BAT_MERCHANT + " = '" +
         this.objFrameworkBatch.getMerchantName() +
         "' AND " +
         BAT_BATCH    + " = '" +
         this.objFrameworkBatch.getBatchNumber() + "'";

      // Construct a JDBC connection.
      ETillArchive varDatabase = new ETillArchive();

      // Lock the JDBC connection.
      synchronized (ETillArchive.getArchiveLock()) {
         try {
            // Trace database read.
            if (Trace.isAnyoneTracing()) {
               Trace.traceDatabaseRead(
                  TRACE_ID,
                  "\n" + varSelect
               );
            }
            // Execute the query.
            ResultSet varResult = varDatabase.beginQuery(varSelect);

            if (varResult.next()) {
               // The required data was found.  Initialize this
               // object's instance data.
               this.setApproval(
                  ETillArchive.readString(varResult, BAT_APPROVAL)
               );
               this.setReason(
                  ETillArchive.readString(varResult, BAT_REASON)
               );
               this.setClosedToTrans(
                  ETillArchive.readInteger(varResult, BAT_CLOSED_TO_TRANS)
               );
               // Assertion test:  If there is more than one row in the
               // result set, there is a serious defect and the
               // cassette cannot continue.  There should never be more
               // than one row for a given cassette batch object.
               // The error is logged and a cassette exception is thrown.
               if (varResult.next()) {
                  if (Trace.isAnyoneTracing()) {
                     Trace.traceErrorOccurred(
                        TRACE_ID,
                        "LdbCardBatch Found too many rows." +
                        "\n\tPRC = " + PRC_DATABASE_ERROR +
                        "\n\tSRC = " + SRC_BATCH_SELECT_ROW_FAILURE
                     );
                  }
                  Vector varVector = new Vector();
                  varVector.addElement(
                     this.objFrameworkBatch.getMerchantName()
                  );
                  varVector.addElement(
                     this.objFrameworkBatch.getBatchNumber()
                  );
                  ErrorLog.log(
                     RB_ID,
                     MSG_BATCH_SELECT_ROW_FAILURE,
                     varVector
                  );
                  throw new ETillCassetteException(
                     ErrorLog.lookupMessage(
                        RB_ID,
                        MSG_BATCH_SELECT_ROW_FAILURE,
                        varVector
                     )
                  );
               }
            } else {
               // The data was not found.  Log an error and abort.
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardBatch Batch resurrection failed." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_BATCH_SELECT_BATCH_MISSING
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_BATCH_SELECT_BATCH_MISSING,
                  null,
                  this.objFrameworkBatch.getMerchantName(),
                  this.objFrameworkBatch.getBatchNumber()
               );
               throw new ETillAbortOperation(
                  PRC_DATABASE_ERROR,
                  SRC_BATCH_SELECT_BATCH_MISSING
               );
            }
         } catch (SQLException varException) {
            // The query failed.  Log the error and abort.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch SQLException caught." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_BATCH_SELECT_SQL_FAILURE
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_BATCH_SELECT_SQL_FAILURE,
               varException,
               this.objFrameworkBatch.getMerchantName(),
               this.objFrameworkBatch.getBatchNumber()
            );
            throw new ETillAbortOperation(
               PRC_DATABASE_ERROR,
               SRC_BATCH_SELECT_SQL_FAILURE
            );
         } finally {
            // Regardless of whether there was an exception or not,
            // clean up the ETillArchive object.
            try {
               varDatabase.endQuery();
            } catch (SQLException varException) {
               // An exception was thrown during cleanup so just
               // log the error and continue.
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardBatch SQL Exception caught." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_BATCH_SELECT_CLOSE_FAILURE
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_BATCH_SELECT_CLOSE_FAILURE,
                  varException,
                  this.objFrameworkBatch.getMerchantName(),
                  this.objFrameworkBatch.getBatchNumber()
               );
            }
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.resurrectBatch()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette batch object on
   * the framework order's create list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void createRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.createRecord()"
         );
      }

      try {
         // The SQL statement to insert a new cassette batch.
         String varInsert =
            " INSERT INTO " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_BATCH + " (" +
            BAT_MERCHANT     + ", " +
            BAT_BATCH        + ", " +
            BAT_APPROVAL     + ", " +
            BAT_REASON       + ", " +
            BAT_CLOSED_TO_TRANS +
            ") VALUES (?, ?, ?, ?, ?)";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varInsert + "\n" +
               BAT_MERCHANT  + " = " +
               this.objFrameworkBatch.getMerchantName() + "\n" +
               BAT_BATCH     + " = " +
               this.objFrameworkBatch.getBatchNumber() + "\n" +
               BAT_APPROVAL  + " = " + this.getApproval() + "\n" +
               BAT_REASON    + " = " + this.getReason() + "\n" +
               BAT_CLOSED_TO_TRANS + " = " + this.getClosedToTrans() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varInsert);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkBatch.getMerchantName(),    varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkBatch.getBatchNumber(),     varStatement, 2
         );
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 3
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 4
         );
         ETillArchive.addInteger(
            this.getClosedToTrans(),                     varStatement, 5
         );
         // Execute the insert.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Created too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_BATCH_CREATE_SQL_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkBatch.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkBatch.getBatchNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_BATCH_CREATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_BATCH_CREATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The insert failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_BATCH_CREATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_BATCH_CREATE_SQL_FAILURE,
            varException,
            this.objFrameworkBatch.getMerchantName(),
            this.objFrameworkBatch.getBatchNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_BATCH_CREATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.createRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette batch object on
   * the framework order's update list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void updateRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.updateRecord()"
         );
      }

      try {
         // The SQL statement to update an existing cassette batch.
         String varUpdate =
            " UPDATE " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_BATCH + " SET " +
            BAT_APPROVAL     + "= ?, " +
            BAT_REASON       + "= ?, "  +
            BAT_CLOSED_TO_TRANS + "= ? "  +
            " WHERE " +
            BAT_MERCHANT + " = ? AND " +
            BAT_BATCH    + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varUpdate + "\n" +
               BAT_APPROVAL  + " = " + this.getApproval() + "\n" +
               BAT_REASON    + " = " + this.getReason() + "\n" +
               BAT_CLOSED_TO_TRANS + " = " +
               this.getClosedToTrans() + "\n" +
               BAT_MERCHANT  + " = " +
               this.objFrameworkBatch.getMerchantName() + "\n" +
               BAT_BATCH     + " = " +
               this.objFrameworkBatch.getBatchNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varUpdate);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 1
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 2
         );
         ETillArchive.addInteger(
            this.getClosedToTrans(),                     varStatement, 3
         );
         ETillArchive.addString(
            this.objFrameworkBatch.getMerchantName(),    varStatement, 4
         );
         ETillArchive.addString(
            this.objFrameworkBatch.getBatchNumber(),     varStatement, 5
         );
         // Execute the update.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Modified too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_BATCH_UPDATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkBatch.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkBatch.getBatchNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_BATCH_UPDATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_BATCH_UPDATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The update failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_BATCH_UPDATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_BATCH_UPDATE_SQL_FAILURE,
            varException,
            this.objFrameworkBatch.getMerchantName(),
            this.objFrameworkBatch.getBatchNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_BATCH_UPDATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.updateRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette batch object on
   * the framework order's delete list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deleteRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.deleteRecord()"
         );
      }

      try {
         // The SQL statement to delete an existing cassette batch.
         String varDelete =
            " DELETE FROM " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_BATCH +
            " WHERE " +
            BAT_MERCHANT + " = ? AND " +
            BAT_BATCH    + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varDelete + "\n" +
               BAT_MERCHANT + " = " +
               this.objFrameworkBatch.getMerchantName() + "\n" +
               BAT_BATCH    + " = " +
               this.objFrameworkBatch.getBatchNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varDelete);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkBatch.getMerchantName(),    varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkBatch.getBatchNumber(),     varStatement, 2
         );
         // Execute the delete.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Deleted too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_BATCH_DELETE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkBatch.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkBatch.getBatchNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_BATCH_DELETE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_BATCH_DELETE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The delete failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_BATCH_DELETE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_BATCH_DELETE_SQL_FAILURE,
            varException,
            this.objFrameworkBatch.getMerchantName(),
            this.objFrameworkBatch.getBatchNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_BATCH_DELETE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.deleteRecord()"
         );
      }
   }

   /*******************************************************************
    * This method is used by the cassette when it is asked to service
    * a BatchOpen request.
    * <p>
    * For this cassette, the payment engine does not allow a merchant to
    * open a batch.  Batches are always implictly opened as needed.
    * The operation is always aborted.
    *
    * @param     argRequest - A BatchOpen batch request.
    * @param     argResponse - The response object to hold the result.
    * @exception ETillAbortOperation - any failure results in this
    * exception.
    *******************************************************************/
   public void batchOpen (
      BatchOpenRequest argRequest,
      APIResponse      argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.batchOpen()"
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceErrorOccurred(
            TRACE_ID,
            "LdbCardBatch BatchOpen not allowed." +
            "\n\tPRC = " + PRC_COMMAND_NOT_SUPPORTED +
            "\n\tSRC = " + RC_NONE
         );
      }
      throw new ETillAbortOperation (
         PRC_COMMAND_NOT_SUPPORTED,
         RC_NONE
      );

   }

   /********************************************************************
   * This method is used by the cassette when it is asked to service
   * an BatchClose request.
   * <p>
   * For this cassette, the account object is used to settle the batch.
   * If successful, the batch will be closed with a status of balanced.
   * Otherwise, the batch will remain open with a status of "out of
   * balance".
   *
   * @param     argRequest - A BatchClose batch request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void batchClose(
      BatchCloseRequest argRequest,
      APIResponse       argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.batchClose()"
         );
      }

      // If the request's connection is null, this request was
      // submitted by an automatic batch close function, not by a
      // merchant via the API.  In this case, it is not considered an
      // error if the batch is already closed.  It simply means that
      // this automatic batch close request lost a race with some
      // other (merchant or automatic) batch close request. When
      // this case occurs, simply set the return codes and return.
      if (
         argRequest.getConnection() == null &&
         this.objFrameworkBatch.getCurrentState() == BATCH_CLOSED
      ) {
         // Set the proper return codes.
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );
      } else {
         // Abort BatchClose if the batch is not open.
         if (this.objFrameworkBatch.getCurrentState() != BATCH_OPEN) {

            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch BatchClose was requested but the " +
                  " batch is not in BATCH_OPEN state" +
                  "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
                  "\n\tSRC    = " + RC_BATCH +
                  "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
               );
            }
            throw new ETillAbortOperation(
               PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
               RC_BATCH
            );

         }

         // Ask the account to settle this batch.  If it returns
         // True, the batch settled successfully.  Otherwise, the
         // settlement failed.  The account will set this
         // object's approval code and decline reason before it returns.
         if (this.objCassetteAccount.settleBatch(this)) {
            // The batch settled.  Set the new state to Closed and the
            // status to Balanced.
            this.objFrameworkBatch.setCurrentState(BATCH_CLOSED);
            this.objFrameworkBatch.setPurgeAllowed(false);
            this.objFrameworkBatch.setBatchStatus(BATCH_BALANCED);
            this.objFrameworkBatch.setTimeClosed(
               Supervisor.getTimestamp(System.currentTimeMillis())
            );
            // For consistency, record the fact that this batch is
            // closed to new transactions, even though this is obvious
            // from the framework batch current state.
            this.closeBatchToNewTrans();
            // Add the framework batch to the update list.
            Supervisor.getThreadCommitPoint().
               addToUpdateList(this.objFrameworkBatch);
            // Ask the Supervisor to remove the old batch from its collection.
            Supervisor.removeBatch(this.objFrameworkBatch);
            // Close all the payments in this batch.
            this.closePayments();
            // Close all the credits in this batch.
            this.closeCredits();

            // Set the proper return codes.
            argResponse.setReturnCodes(
               PRC_OPERATION_SUCCESS,
               RC_NONE
            );
         } else {
            // The settlement failed.  Log the error and abort.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Batch settlement failed." +
                  "\n\tPRC    = " + PRC_BATCH_ERROR +
                  "\n\tSRC    = " + RC_CASSETTE_BATCH_ERROR +
                  "\n\tReason = " + this.getReason()
               );
            }
            // Update the status.
            this.objFrameworkBatch.setBatchStatus(BATCH_OUT_OF_BALANCE);
            // Even though the batch state goes back to open, the
            // account object should no longer select this batch for
            // new transactions.  Adding new payments or credits to
            // a batch that is out of balance won't help and may
            // harm the batch.  Therefore, set the indicator.
            this.closeBatchToNewTrans();
            // Add the framework batch to the update list.
            Supervisor.getThreadCommitPoint().
               addToUpdateList(this.objFrameworkBatch);
            // Set the proper return codes.
            argResponse.setReturnCodes(
               PRC_BATCH_ERROR,
               RC_CASSETTE_BATCH_ERROR
            );
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.batchClose()"
         );
      }
   }

   /********************************************************************
   * This method is used to close all the payments in a closing batch.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void closePayments()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.closePayments()"
         );
      }

      // Obtain an enumeration of keys for all the payments in the batch.
      Enumeration varKeyList = this.objFrameworkBatch.getPaymentList();
      while (varKeyList.hasMoreElements()) {

         TransactionKey varKey = (TransactionKey)varKeyList.nextElement();
         Payment varFrameworkPayment = this.retrieveBatchPayment(varKey);
         Order varFrameworkOrder = varFrameworkPayment.getOrder();
         LdbCardOrder varCassetteOrder =
            (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
         varCassetteOrder.closePayment(varFrameworkPayment);

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.closePayments()"
         );
      }
   }

   /********************************************************************
   * This method is used to close all the credits in a closing batch.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private void closeCredits()
      throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.closeCredits()"
         );
      }

      // Obtain an enumeration of keys for all the credits in the batch.
      Enumeration varKeyList = this.objFrameworkBatch.getCreditList();
      while (varKeyList.hasMoreElements()) {
         TransactionKey varKey = (TransactionKey)varKeyList.nextElement();

         Credit varFrameworkCredit = this.retrieveBatchCredit(varKey);
         Order varFrameworkOrder = varFrameworkCredit.getOrder();
         LdbCardOrder varCassetteOrder =
            (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
         varCassetteOrder.closeCredit(varFrameworkCredit);


      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.closeCredits()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette when it is asked to service
   * a BatchPurge request.
   * <p>
   * For this cassette, the method attempts to reverse every deposit and
   * every refund in the open batch.  Note that an exception thrown by
   * either case is not considered fatal.  A purge operation, by
   * definition, does the best it can to empty the batch both locally
   * and in the financial network.  If any payment or credit cannot
   * be removed, the purge operation continues.  If this type of error
   * occurs, human intervention may be required to synchronize the
   * Payment Manager with the financial network.
   *
   * @param     argRequest - A BatchPurge batch request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void batchPurge(
      BatchPurgeRequest argRequest,
      APIResponse       argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.batchPurge()"
         );
      }

      // Abort BatchPurge if the batch is not in OPEN state.
      if (this.objFrameworkBatch.getCurrentState() != BATCH_OPEN) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch BatchPurge was requested but the " +
               " batch is not in BATCH_OPEN state" +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_BATCH +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_BATCH
         );
      }

      // Purge all the payments in this batch.
      boolean varPurgeIncomplete = this.purgePayments(
         argRequest,
         argResponse
      );
      // Purge all the credits in this batch.
      boolean varCreditIncomplete = this.purgeCredits(
         argRequest,
         argResponse
      );

      // Add the framework batch to the update list.
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkBatch);
      // Set the proper return codes.
      short varSecondaryReturnCode = RC_NONE;
      if (varPurgeIncomplete || varCreditIncomplete) {
         varSecondaryReturnCode = SRC_BATCH_PURGE_INCOMPLETE;
      }
      argResponse.setReturnCodes(
         PRC_OPERATION_SUCCESS,
         varSecondaryReturnCode
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.batchPurge()"
         );
      }
   }

   /********************************************************************
   * This method is used to purge all the payments in an open batch.
   *
   * @param     argRequest - A BatchPurge batch request.
   * @param     argResponse - The response object to hold the result.
   * @return    - An incomplete indicator: true if some payments could
   * not be removed from the batch, false otherwise.
   ********************************************************************/
   private boolean purgePayments(
      BatchPurgeRequest argRequest,
      APIResponse       argResponse
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.purgePayments()"
         );
      }

      // Assume all payments will be removed from the batch.
      boolean varIncomplete = false;

      // Obtain an enumeration of keys for all the payments in the batch.
      Enumeration varKeyList = this.objFrameworkBatch.getPaymentList();
      while (varKeyList.hasMoreElements()) {
         TransactionKey varKey = (TransactionKey)varKeyList.nextElement();
         try {
            Payment varFrameworkPayment = this.retrieveBatchPayment(varKey);
            Order varFrameworkOrder = varFrameworkPayment.getOrder();
            LdbCardOrder varCassetteOrder =
               (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
            try {
               DepositReversalRequest varRequest = new DepositReversalRequest(
                  argRequest.getConnection(),
                  varFrameworkPayment,
                  argRequest.getFrameworkKeywords(),
                  argRequest.getProtocolData()
               );
               varCassetteOrder.depositReversal(
                  varRequest,
                  argResponse
               );
               // Don't stop the purge just because a reversal failed.
               // Instead, trace and log the error. Then continue.
               if (argResponse.getPrimaryRC() != PRC_OPERATION_SUCCESS) {
                  varIncomplete = true;
                  if (Trace.isAnyoneTracing()) {
                     Trace.traceErrorOccurred(
                        TRACE_ID,
                        "LdbCardBatch Error reversing a deposit" +
                        " during a purge." +
                        "\n\tPRC = " + argResponse.getPrimaryRC() +
                        "\n\tSRC = " + argResponse.getSecondaryRC()
                     );
                  }
                  Vector varVector = new Vector();
                  varVector.addElement(
                     this.objFrameworkBatch.getMerchantName()
                  );
                  varVector.addElement(
                     this.objFrameworkBatch.getAccountNumber()
                  );
                  varVector.addElement(
                     varKey.getOrderNumber()
                  );
                  varVector.addElement(
                     varKey.getTransactionNumber()
                  );
                  varVector.addElement(
                     String.valueOf(argResponse.getPrimaryRC())
                  );
                  varVector.addElement(
                     String.valueOf(argResponse.getSecondaryRC())
                  );
                  ErrorLog.log(
                     RB_ID,
                     MSG_BATCH_DEPOSIT_REVERSAL_FAILURE,
                     varVector
                  );
               }
            } catch (ETillAbortOperation varException) {
               // Don't stop the purge just because a reversal failed.
               // Instead, trace and log the error.  Then continue.
               varIncomplete = true;
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardBatch Error reversing a deposit during a purge." +
                     "\n\tPRC = " + varException.getPrimaryRC() +
                     "\n\tSRC = " + varException.getSecondaryRC()
                  );
               }
               Vector varVector = new Vector();
               varVector.addElement(
                  this.objFrameworkBatch.getMerchantName()
               );
               varVector.addElement(
                  this.objFrameworkBatch.getAccountNumber()
               );
               varVector.addElement(
                  varKey.getOrderNumber()
               );
               varVector.addElement(
                  varKey.getTransactionNumber()
               );
               varVector.addElement(
                  String.valueOf(varException.getPrimaryRC())
               );
               varVector.addElement(
                  String.valueOf(varException.getSecondaryRC())
               );
               ErrorLog.log(
                  RB_ID,
                  MSG_BATCH_DEPOSIT_REVERSAL_FAILURE,
                  varVector
               );
               ErrorLog.printStackTrace(varException);
            }
         } catch (ETillAbortOperation varException) {
            // Don't stop the purge just because the Supervisor could
            // not retrieve an order from the database.  Instead,
            // trace and log the error.  Then continue.
            varIncomplete = true;
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Supervisor could not retrieve order." +
                  "\n\tPRC = " + varException.getPrimaryRC() +
                  "\n\tSRC = " + varException.getSecondaryRC()
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkBatch.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkBatch.getAccountNumber()
            );
            varVector.addElement(
               varKey.getOrderNumber()
            );
            varVector.addElement(
               varKey.getTransactionNumber()
            );
            varVector.addElement(
               String.valueOf(varException.getPrimaryRC())
            );
            varVector.addElement(
               String.valueOf(varException.getSecondaryRC())
            );
            ErrorLog.log(
               RB_ID,
               MSG_BATCH_CREDIT_RETRIEVAL_FAILURE,
               varVector
            );
            ErrorLog.printStackTrace(varException);
         }

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.purgePayments()"
         );
      }
      return varIncomplete;
   }

   /********************************************************************
   * This method is used to purge all the credits in an open batch.
   *
   * @param     argRequest - A BatchPurge batch request.
   * @param     argResponse - The response object to hold the result.
   * @return    - An incomplete indicator: true if some credits could
   * not be removed from the batch, false otherwise.
   ********************************************************************/
   private boolean purgeCredits(
      BatchPurgeRequest argRequest,
      APIResponse       argResponse
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.purgeCredits()"
         );
      }

      // Assume all credits will be removed from the batch.
      boolean varIncomplete = false;

      // Obtain an enumeration of keys for all the credits in the batch.
      Enumeration varKeyList = this.objFrameworkBatch.getCreditList();
      while (varKeyList.hasMoreElements()) {
         TransactionKey varKey = (TransactionKey)varKeyList.nextElement();
         try {
            Credit varFrameworkCredit = this.retrieveBatchCredit(varKey);
            Order varFrameworkOrder = varFrameworkCredit.getOrder();
            LdbCardOrder varCassetteOrder =
               (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
            try {
               RefundReversalRequest varRequest = new RefundReversalRequest(
                  argRequest.getConnection(),
                  varFrameworkCredit,
                  argRequest.getFrameworkKeywords(),
                  argRequest.getProtocolData()
               );
               varCassetteOrder.refundReversal(
                  varRequest,
                  argResponse
               );
               // Don't stop the purge just because a reversal failed.
               // Instead, trace and log the error. Then continue.
               if (argResponse.getPrimaryRC() != PRC_OPERATION_SUCCESS) {
                  varIncomplete = true;
                  if (Trace.isAnyoneTracing()) {
                     Trace.traceErrorOccurred(
                        TRACE_ID,
                        "LdbCardBatch Error reversing a credit" +
                        " during a purge." +
                        "\n\tPRC = " + argResponse.getPrimaryRC() +
                        "\n\tSRC = " + argResponse.getSecondaryRC()
                     );
                  }
                  Vector varVector = new Vector();
                  varVector.addElement(
                     this.objFrameworkBatch.getMerchantName()
                  );
                  varVector.addElement(
                     this.objFrameworkBatch.getAccountNumber()
                  );
                  varVector.addElement(
                     varKey.getOrderNumber()
                  );
                  varVector.addElement(
                     varKey.getTransactionNumber()
                  );
                  varVector.addElement(
                     String.valueOf(argResponse.getPrimaryRC())
                  );
                  varVector.addElement(
                     String.valueOf(argResponse.getSecondaryRC())
                  );
                  ErrorLog.log(
                     RB_ID,
                     MSG_BATCH_REFUND_REVERSAL_FAILURE,
                     varVector
                  );
               }
            } catch (ETillAbortOperation varException) {
               // Don't stop the purge just because a reversal failed.
               // Instead, trace and log the error.  Then continue.
               varIncomplete = true;
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardBatch Error reversing a credit during a purge." +
                     "\n\tPRC = " + varException.getPrimaryRC() +
                     "\n\tSRC = " + varException.getSecondaryRC()
                  );
               }
               Vector varVector = new Vector();
               varVector.addElement(
                  this.objFrameworkBatch.getMerchantName()
               );
               varVector.addElement(
                  this.objFrameworkBatch.getAccountNumber()
               );
               varVector.addElement(
                  varKey.getOrderNumber()
               );
               varVector.addElement(
                  varKey.getTransactionNumber()
               );
               varVector.addElement(
                  String.valueOf(varException.getPrimaryRC())
               );
               varVector.addElement(
                  String.valueOf(varException.getSecondaryRC())
               );
               ErrorLog.log(
                  RB_ID,
                  MSG_BATCH_REFUND_REVERSAL_FAILURE,
                  varVector
               );
               ErrorLog.printStackTrace(varException);
            }
         } catch (ETillAbortOperation varException) {
            // Don't stop the purge just because the Supervisor could
            // not retrieve an order from the database.  Instead,
            // trace and log the error.  Then continue.
            varIncomplete = true;
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Supervisor could not retrieve order." +
                  "\n\tPRC = " + varException.getPrimaryRC() +
                  "\n\tSRC = " + varException.getSecondaryRC()
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkBatch.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkBatch.getAccountNumber()
            );
            varVector.addElement(
               varKey.getOrderNumber()
            );
            varVector.addElement(
               varKey.getTransactionNumber()
            );
            varVector.addElement(
               String.valueOf(varException.getPrimaryRC())
            );
            varVector.addElement(
               String.valueOf(varException.getSecondaryRC())
            );
            ErrorLog.log(
               RB_ID,
               MSG_BATCH_CREDIT_RETRIEVAL_FAILURE,
               varVector
            );
            ErrorLog.printStackTrace(varException);
         }

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.purgeCredits()"
         );
      }
      return varIncomplete;
   }

   /********************************************************************
   * This method is used by the batch object to retrieve a payment and
   * its order into the cache.  Any retrieved objects will be released
   * from the cache automatically when the current commit point is
   * committed or rejected.
   *
   * @param     argKey - The transaction key taken from the framework
   * batch payment list.
   * @return    - A framework payment object for the specified key.
   * @exception ETillAbortOperation - any error will result in this
   * exception.
   ********************************************************************/
   public Payment retrieveBatchPayment(
      TransactionKey argKey
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.retrieveBatchPayment()"
         );
      }
      Order varFrameworkOrder =
         Supervisor.retrieveOrder(
            this.objFrameworkBatch.getMerchantName(),
            argKey.getOrderNumber()
         );
      // Note: If there is a database error, an ETillAbortException
      // will be thrown and the following assertion test will not
      // be executed.
      // Assertion test:  If order is null, there is a serious defect
      // and the cassette cannot continue.
      if (varFrameworkOrder == null) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Payment's order cannot be retrieved." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_NULL_ORDER_FOR_PAYMENT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_NULL_ORDER_FOR_PAYMENT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_NULL_ORDER_FOR_PAYMENT,
               varVector
            )
         );

      }

      Payment varFrameworkPayment =
         varFrameworkOrder.getPayment(
            argKey.getTransactionNumber()
         );
      // Assertion test: If the payment is null, or if the payment's
      // batch number does not match this batch number there is a
      // serious defect and the cassette cannot continue.
      if (varFrameworkPayment == null) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Payment cannot be retrieved." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_NULL_PAYMENT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_NULL_PAYMENT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_NULL_PAYMENT,
               varVector
            )
         );
      } else if (
         ! varFrameworkPayment.getBatchNumber().equals(
            this.objFrameworkBatch.getBatchNumber()
         )
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Payment contains different batch number." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_BAD_BATCH_IN_PAYMENT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_BAD_BATCH_IN_PAYMENT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_BAD_BATCH_IN_PAYMENT,
               varVector
            )
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.retrieveBatchPayment()"
         );
      }
      return varFrameworkPayment;
   }

   /********************************************************************
   * This method is used by the batch object to retrieve a credit and
   * its order into the cache.  Any retrieved objects will be released
   * from the cache automatically when the current commit point is
   * committed or rejected.
   *
   * @param     argKey - The transaction key taken from the framework
   * batch credit list.
   * @return    - A framework credit object for the specified key.
   * @exception ETillAbortOperation - any error will result in this
   * exception.
   ********************************************************************/
   public Credit retrieveBatchCredit(
      TransactionKey argKey
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.retrieveBatchCredit()"
         );
      }

      Order varFrameworkOrder =
         Supervisor.retrieveOrder(
            this.objFrameworkBatch.getMerchantName(),
            argKey.getOrderNumber()
         );
      // Note: If there is a database error, an ETillAbortException
      // will be thrown and the following assertion test will not
      // be executed.
      // Assertion test:  If order is null, there is a serious defect
      // and the cassette cannot continue.
      if (varFrameworkOrder == null) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Credit's order cannot be retrieved." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_NULL_ORDER_FOR_CREDIT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_NULL_ORDER_FOR_CREDIT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_NULL_ORDER_FOR_CREDIT,
               varVector
            )
         );
      }

      Credit varFrameworkCredit =
         varFrameworkOrder.getCredit(
            argKey.getTransactionNumber()
         );
      // Assertion test: If the credit is null, or if the credit's
      // batch number does not match this batch number there is a
      // serious defect and the cassette cannot continue.
      if (varFrameworkCredit == null) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Credit cannot be retrieved." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_NULL_CREDIT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_NULL_CREDIT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_NULL_CREDIT,
               varVector
            )
         );
      } else if (
         ! varFrameworkCredit.getBatchNumber().equals(
            this.objFrameworkBatch.getBatchNumber()
         )
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Credit contains different batch number." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_BAD_BATCH_IN_CREDIT
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argKey.getOrderNumber()
         );
         varVector.addElement(
            argKey.getTransactionNumber()
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_BAD_BATCH_IN_CREDIT,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_BAD_BATCH_IN_CREDIT,
               varVector
            )
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.retrieveBatchCredit()"
         );
      }
      return varFrameworkCredit;
   }

   /********************************************************************
   * This method is used by the cassette when it is asked to service
   * an DeleteBatch request.
   * <p>
   *
   * @param     argRequest - A DeleteBatch batch request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void batchDelete(
      DeleteBatchRequest argRequest,
      APIResponse       argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.deleteBatch()"
         );
      }

      // Abort deleteBatch if the batch is not in CLOSED state.
      if (this.objFrameworkBatch.getCurrentState() != BATCH_CLOSED) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch BatchDelete was requested but the " +
               " batch is not in BATCH_CLOSED state" +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_BATCH +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_BATCH
         );
      }

      // Check for payments or credits in this batch.  If there are
      // none, perform the delete operation. Otherwise abort.
      if (this.isDeleteBatchAllowed()) {

         // Ask the supervisor to remove the batch;
         Supervisor.removeBatch(this.objFrameworkBatch);
         // Add the framework batch to the update list.
         Supervisor.getThreadCommitPoint().
            addToDeleteList(this.objFrameworkBatch);
         // Set the proper return codes.
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );

      } else {

         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch BatchDelete was requested but it " +
               " still contains payments or credits." +
               "\n\tPRC    = " + PRC_DELETE_ERROR +
               "\n\tSRC    = " + RC_BATCH +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_DELETE_ERROR,
            RC_BATCH
         );

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.deleteBatch()"
         );
      }
   }

   /********************************************************************
   * This method is used to determine if it is safe to delete a batch.
   *
   * @return    True if the payment and credit list is empty, false
   * otherwise.
   ********************************************************************/
   private boolean isDeleteBatchAllowed() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.isDeleteBatchAllowed()"
         );
      }

      boolean varResult = true;

      Enumeration varPaymentList = this.objFrameworkBatch.getPaymentList();
      Enumeration varCreditList = this.objFrameworkBatch.getCreditList();

      if (varPaymentList.hasMoreElements()) {
         varResult = false;
      } else if (varCreditList.hasMoreElements()) {
         varResult = false;
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.isDeleteBatchAllowed()"
         );
      }

      return varResult;
   }

   /********************************************************************
   * This method is used to add a payment to a batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette batch object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkPayment - the framework payment object to be
   * removed from the batch.
   ********************************************************************/
   public synchronized void addPayment(
      Payment argFrameworkPayment
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.addPayment()"
         );
      }

      // Assertion Test:  Check to see if the payment can be added to
      // this batch.
      if (
         this.objFrameworkBatch.getCurrentState() != BATCH_OPEN ||
         this.isBatchClosedToNewTransactions()
      ) {
         // This should never happen.  The logic that decided that
         // this batch is to hold the payment identified by a Deposit
         // request should not have chosen this batch unless it is
         // in the BATCH_OPEN state and open to new transactions.  In
         // addition, that logic should be synchronized while testing
         // these conditions and calling addPayment() or addCredit()
         // methods to prevent one thread from chosing this batch while
         // another thread is about to trigger an automatic batch close.
         //
         // If this code is reached, there is a serious defect in the
         // logic that lead up to the call to this method.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Caller attempted to add a payment but" +
               " either the batch is not open or has been closed to" +
               " new transactions." +
               "\n\tPRC    = " + PRC_CASSETTE_ERROR +
               "\n\tSRC    = " + SRC_BATCH_BAD_ADD_PAYMENT_STATE +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState() +
               "\n\tClose Flag = " + this.getClosedToTrans()
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argFrameworkPayment.getOrderNumber()
         );
         varVector.addElement(
            argFrameworkPayment.getTransactionNumber()
         );
         varVector.addElement(
            Integer.toString(
               this.objFrameworkBatch.getCurrentState()
            )
         );
         varVector.addElement(
            this.getClosedToTrans().toString()
         );
         varVector.addElement(
            Integer.toString(PRC_CASSETTE_ERROR)
         );
         varVector.addElement(
            Integer.toString(SRC_BATCH_BAD_ADD_PAYMENT_STATE)
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_BAD_ADD_PAYMENT_STATE,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_BAD_ADD_PAYMENT_STATE,
               varVector
            )
         );
      }

      // Add the payment to the batch.
      this.objFrameworkBatch.addPayment(argFrameworkPayment);

      // Check if an automatic close should be performed and, if so,
      // change the state of the batch and schedule a close.
      this.scheduleAutoCloseIfRequired();

      // In any case, add the framework batch to the update list.
      Supervisor.getThreadCommitPoint().addToUpdateList(
         this.objFrameworkBatch
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.addPayment()"
         );
      }
   }

   /********************************************************************
   * This method is used to remove a payment from a batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette batch object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkPayment - the framework payment object to be
   * removed from the batch.
   * @exception ETillAbortOperation - any error results in this
   * exception.
   ********************************************************************/
   public synchronized void removePayment(
      Payment argFrameworkPayment
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.addPayment()"
         );
      }

      // A batch state check is required.  Even if the batch state
      // is BATCH_OPEN, the batch may be closed because an automatic
      // close was scheduled.  Note that if the current batch state
      // is BATCH_CLOSED, all its payments would be closed and the
      // DepositReversal would have been aborted by the state checks
      // in the cassette payment object.
      if (
         this.objFrameworkBatch.getCurrentState() != BATCH_OPEN ||
         this.isBatchClosedForAutomaticClose()
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch removePayment was requested but the " +
               " batch is not in BATCH_OPEN state" +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_BATCH +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_BATCH
         );
      }

      // Ask the framework batch object to remove the payment.
      this.objFrameworkBatch.removePayment(argFrameworkPayment);
      // Add the framework batch object to the update list.
      Supervisor.getThreadCommitPoint().addToUpdateList(
         this.objFrameworkBatch
      );


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.addPayment()"
         );
      }
   }

   /********************************************************************
   * This method is used to add a credit to a batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette batch object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkCredit - the framework payment credit to be
   * removed from the batch.
   ********************************************************************/
   public synchronized void addCredit(
      Credit argFrameworkCredit
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.addCredit()"
         );
      }

      // Assertion Test: Check to see if the credit can be added to
      // this batch.
      if (
         this.objFrameworkBatch.getCurrentState() != BATCH_OPEN ||
         this.isBatchClosedToNewTransactions()
      ) {
         // This should never happen.  The logic that decided that
         // this batch is to hold the credit identified by a Refund
         // request should not have chosen this batch unless it is
         // in the BATCH_OPEN state and open to new transactions.  In
         // addition, that logic should be synchronized while testing
         // these conditions and calling addPayment() or addCredit()
         // methods to prevent one thread from chosing this batch while
         // another thread is about to trigger an automatic batch close.
         //
         // If this code is reached, there is a serious defect in the
         // logic that lead up to the call to this method.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Caller attempted to add a credit but" +
               " either the batch is not open or has been closed to" +
               " new transactions." +
               "\n\tPRC    = " + PRC_CASSETTE_ERROR +
               "\n\tSRC    = " + SRC_BATCH_BAD_ADD_CREDIT_STATE +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState() +
               "\n\tClose Flag = " + this.getClosedToTrans()
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkBatch.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkBatch.getBatchNumber()
         );
         varVector.addElement(
            argFrameworkCredit.getOrderNumber()
         );
         varVector.addElement(
            argFrameworkCredit.getTransactionNumber()
         );
         varVector.addElement(
            Integer.toString(
               this.objFrameworkBatch.getCurrentState()
            )
         );
         varVector.addElement(
            this.getClosedToTrans().toString()
         );
         varVector.addElement(
            Integer.toString(PRC_CASSETTE_ERROR)
         );
         varVector.addElement(
            Integer.toString(SRC_BATCH_BAD_ADD_CREDIT_STATE)
         );
         ErrorLog.log(
            RB_ID,
            MSG_BATCH_BAD_ADD_CREDIT_STATE,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_BATCH_BAD_ADD_CREDIT_STATE,
               varVector
            )
         );
      }

      // Add the credit to the batch.
      this.objFrameworkBatch.addCredit(argFrameworkCredit);

      // Check if an automatic close should be performed and, if so,
      // change the state of the batch and schedule a close.
      this.scheduleAutoCloseIfRequired();

      // In any case, add the framework batch to the update list.
      Supervisor.getThreadCommitPoint().addToUpdateList(
         this.objFrameworkBatch
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.addCredit()"
         );
      }
   }

   /********************************************************************
   * This method is used to remove a credit from a batch.
   * <p>
   * The method must be synchronized because framework API locks do not
   * prevent race conditions between automatic batch close requests
   * and API requests that affect the contents of the batch (Deposit,
   * DepositReversal, Refund and RefundReversal).  Synchronizing on the
   * cassette batch object prevents one thread from changing the
   * contents of the batch a micro-moment after some other thread has
   * triggered an automatic batch close.
   *
   * @param     argFrameworkCredit - the framework credit object to be
   * removed from the batch.
   * @exception ETillAbortOperation - any error results in this
   * exception.
   ********************************************************************/
   public synchronized void removeCredit(
      Credit argFrameworkCredit
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.addCredit()"
         );
      }

      // A batch state check is required.  Even if the batch state
      // is BATCH_OPEN, the batch may be closed because an automatic
      // close was scheduled.  Note that if the current batch state
      // is BATCH_CLOSED, all its credits would be closed and the
      // RefundReversal would have been aborted by the state checks
      // in the cassette credit object.
      if (
         this.objFrameworkBatch.getCurrentState() != BATCH_OPEN ||
         this.isBatchClosedForAutomaticClose()
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch removeCredit was requested but the " +
               " batch is not in BATCH_OPEN state" +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_BATCH +
               "\n\tState  = " + this.objFrameworkBatch.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_BATCH
         );
      }

      // Ask the framework batch object to remove the credit.
      this.objFrameworkBatch.removeCredit(argFrameworkCredit);
      // Add the framework batch object to the update list.
      Supervisor.getThreadCommitPoint().addToUpdateList(
         this.objFrameworkBatch
      );


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.addCredit()"
         );
      }
   }

   /********************************************************************
   * This method is used to determine if an automatic time of day batch
   * close should be scheduled and, if so, to schedule the automatic
   * close.
   ********************************************************************/
   private void scheduleTimeOfDayCloseIfRequired(
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.scheduleTimeOfDayCloseIfRequired()"
         );
      }

      // Check to see if the merchant has configured the account to
      // automatically close the batch at a given time of day.
      if (this.objCassetteAccount.isBatchCloseTimeConfigured()) {
         // An automatic time of day close has been requested by the
         // merchant.  Build the appropriate cassette work item and
         // add it to the timer queue.
         try {
            this.objTimerWorkItem = new TimeOfDayWorkItem(
               this.buildBatchCloseRequest(),
               this.cassette()
            );
         } catch (ETillAbortOperation varException) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Unable to schedule an automatic " +
                  "BatchClose command." +
                  "\n\tPRC = " + PRC_CASSETTE_ERROR +
                  "\n\tSRC = " + SRC_BATCH_TIMER_WORK_ITEM_FAILURE
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_BATCH_TIMER_WORK_ITEM_FAILURE,
               varException,
               this.objFrameworkBatch.getMerchantName(),
               this.objFrameworkBatch.getBatchNumber(),
               this.objFrameworkBatch.getAccountNumber()
            );
            // Conventional wisdom declares that failure to
            // schedule an automatic close operation should not
            // abort the current operation (most likely the
            // construction or ressurection of a cassette batch
            // object).  If this design point is incompatible with
            // your cassette, remove the comments from the following
            // statement.
            // throw new ETillAbortOperation(
            //    PRC_CASSETTE_ERROR,
            //    SRC_BATCH_TIMER_WORK_ITEM_FAILURE
            // );
         }

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.scheduleTimeOfDayCloseIfRequired()"
         );
      }
   }

   /********************************************************************
   * This method is used to determine if an automatic batch close
   * should be scheduled and, if so, to schedule the automatic close.
   ********************************************************************/
   private void scheduleAutoCloseIfRequired(
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.scheduleAutoCloseIfRequired()"
         );
      }

      // Check to see if the merchant has configured the account to
      // automatically close a batch when it reaches a maximum size.
      if (this.objCassetteAccount.isMaxBatchSizeConfigured()) {
         // An automatic close will be performed if the number of
         // transactions in the batch have reached the maximum.
         // Obtain a list of all the batch item objects (there is one
         // for each currency supported) and add up the number of
         // payments and credits in each.
         Enumeration varItemList = this.objFrameworkBatch.getTotals();
         int varBatchSize = 0;
         while (varItemList.hasMoreElements()) {
            BatchTotalsItem varItem =
               (BatchTotalsItem)varItemList.nextElement();
            varBatchSize +=
               varItem.getNumberOfCredits() +
               varItem.getNumberOfPayments();
         }
         // Check to see if the number of transactions in the batch
         // has reached the maximum number of transactions allowed.
         if (varBatchSize >= this.objCassetteAccount.getMaxBatchSize()) {

            // The maximum has been reached.  Close the batch.
            this.scheduleAutoClose();

         }

      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.scheduleAutoCloseIfRequired()"
         );
      }
   }

   /********************************************************************
   * This method is used to automatically close this batch.  It
   * constructs a work item that will be processed asynchronously
   * by a service thread.  A service thread is required because the
   * current thread does not hold the locks required to close the
   * batch in a thread safe manner.  The service thread will get
   * the correct locks (as specified by a BatchCloseRequest object
   * constructed here) before calling the work item's doYourWork()
   * method.
   * <p>
   * This method may be used by methods in this object when they
   * discover that the maximum batch size has been reached or by
   * methods in a TimeOfDayWorkItem when an automatic time-of-day
   * batch close is to be performed.
   *
   ********************************************************************/
   public void scheduleAutoClose() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.scheduleAutoClose()"
         );
      }

      try {

         // Mark this batch as closed to new transactions due to an
         // automatic batch close.
         this.closeBatchForAutomaticClose();

         // Construct a work item to close the batch.  The work item
         // automatically adds itself to the service thread queue.
         this.objAutoCloseWorkItem = new AutoCloseWorkItem(
            this.buildBatchCloseRequest(),
            this.cassette()
         );
         // Add the updated framework batch object to the update list.
         Supervisor.getThreadCommitPoint().addToUpdateList(
            this.objFrameworkBatch
         );


      } catch (ETillAbortOperation varException) {

         // An error occured either while constructing the
         // batch close request or while constructing the work item.
         // In either case, log the error and abort the current
         // operation.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardBatch Unable to schedule an automatic " +
               "BatchClose command." +
               "\n\tPRC = " + PRC_CASSETTE_ERROR +
               "\n\tSRC = " + SRC_BATCH_CLOSE_WORK_ITEM_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_BATCH_CLOSE_WORK_ITEM_FAILURE,
            varException,
            this.objFrameworkBatch.getMerchantName(),
            this.objFrameworkBatch.getBatchNumber(),
            this.objFrameworkBatch.getAccountNumber()
         );
         // Conventional wisdom declares that failure to
         // schedule an automatic close operation should not
         // abort the current operation (most likely a deposit
         // or refund request that adds a payment or credit to
         // the framework batch).  If this design point is incompatible
         // with your cassette, remove the comments from the following
         // statement.
         // throw new ETillAbortOperation(
         //    PRC_CASSETTE_ERROR,
         //    SRC_BATCH_CLOSE_WORK_ITEM_FAILURE
         // );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.scheduleAutoClose()"
         );
      }
   }

   /********************************************************************
   * This method is used to manufacture an internal batch close request.
   *
   * @return    - A BatchCloseRequest
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   private BatchCloseRequest buildBatchCloseRequest(
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardBatch.buildBatchCloseRequest()"
         );
      }

      BatchCloseRequest varResult = null;

      // Create a new parameter table and initialize it with the
      // correct parameters, just as if the merchant issued a
      // BatchCloseRequest command.
      ParameterTable varParameters = new ParameterTable();
      varParameters.put(
         COMMAND,
         CLOSE_BATCH
      );
      varParameters.put(
         MERCHANTNUMBER,
         this.objFrameworkBatch.getMerchantName()
      );
      varParameters.put(
         BATCHNUMBER,
         this.objFrameworkBatch.getBatchNumber()
      );
      varParameters.put(
         ETAPIVERSION,
         "3"
      );
      varParameters.put(
         FORCE,
         DO_NOT_ATTEMPT
      );

      // Construct a new request object with the prebuilt parameter
      // table.  Note that the value for the ETillConnection parameter
      // is null.  This is how the cassette can distinguish between
      // a request built by a merchant command and an internal
      // request.  (For commands issued by the merchant, the
      // ETillConnection is required to send the reply back to the
      // merchant.)
      varResult = new BatchCloseRequest(
         null,                         // No ETillConnection
         this.objFrameworkBatch,
         varParameters,
         null                          // No protocol data
      );

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardBatch.buildBatchCloseRequest()"
         );
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to obtain the cassette account object used
   * to service this batch;
   *
   * @return    - The cassette account object used by this batch.
   ********************************************************************/
   public LdbCardAccount getAccount() {
      return this.objCassetteAccount;
   }

   /********************************************************************
   * This method is used to obtain the approval code returned by the
   * payment engine.
   *
   * @return    - The approval code returned by the payment engine.
   ********************************************************************/
   public String getApproval() {
      return this.objApprovalCode;
   }

   /********************************************************************
   * This method is used to obtain the decline reason returned by the
   * payment engine.
   *
   * @return    - The decline reason returned by the payment engine.
   ********************************************************************/
   public String getReason() {
      return this.objDeclineReason;
   }

   /********************************************************************
   * This method is used to obtain an Integer object that indicates
   * whether or not this batch is closed to new transactions.  This is
   * a private method; it is to be used exclusively for getting an
   * object used to update the data base.
   *
   * @return    - An Integer object that indicates whether or not this
   * batch is closed to new transactions.
   ********************************************************************/
   private Integer getClosedToTrans() {
      return this.objClosedToTrans;
   }

   /********************************************************************
   * This method is used to test whether or not this batch is closed
   * to new transactions.
   *
   * @return    - False if the batch can be used for new transactions,
   * true otherwise.
   ********************************************************************/
   public boolean isBatchClosedToNewTransactions() {
      boolean varResult = false;
      if (this.objClosedToTrans.intValue() != 0) {
         varResult = true;
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to test whether or not this batch is closed
   * to new transactions due to a pending automatic batch close.
   *
   * @return    - False if the batch can be used for new transactions,
   * true otherwise.
   ********************************************************************/
   public boolean isBatchClosedForAutomaticClose() {
      boolean varResult = false;
      if (this.objClosedToTrans.intValue() > 1) {
         varResult = true;
      }
      return varResult;
   }

   /********************************************************************
   * This method is used to set the approval code.
   *
   * @param     argApprovalCode - The new approval code returned by
   * the payment engine.
   ********************************************************************/
   public void setApproval(
      String argApprovalCode
   ) {
      this.objApprovalCode = argApprovalCode;
   }

   /********************************************************************
   * This method is used to set the decline reason.
   *
   * @param     argDeclineReason - The new decline reason returned by
   * the payment engine.
   ********************************************************************/
   public void setReason(
      String argDeclineReason
   ) {
      this.objDeclineReason = argDeclineReason;
   }

   /********************************************************************
   * This method is used to set the Integer object that indicates
   * whether or not this batch is closed to new transactions.  This is
   * a private method; it is to be used exclusively for processing an
   * object returned from a database query.
   *
   * @param     argClosedToTrans - The Integer object that indicates
   * whether or not this batch is closed to new transactions.
   ********************************************************************/
   private void setClosedToTrans(
      Integer argClosedToTrans
   ) {
      this.objClosedToTrans = argClosedToTrans;
   }

   /********************************************************************
   * This method is used to set the value that indicates
   * whether or not this batch is closed to new transactions.
   ********************************************************************/
   private void closeBatchToNewTrans(
   ) {
      this.objClosedToTrans = new Integer(1);
   }

   /********************************************************************
   * This method is used to set the value that indicates
   * whether or not this batch is closed due to a pending automatic
   * batch close.
   ********************************************************************/
   private void closeBatchForAutomaticClose(
   ) {
      this.objClosedToTrans = new Integer(2);
   }

   /********************************************************************
   * The AutoCloseWorkItem object is used to automatically close an
   * open batch as soon as possible.  f day (specified in minutes after
   * midnight).  When constructed, it automatically schedules itself
   * using the framework's timer queue.  When the scheduled time is
   * reached, the framework gives this object to the next available
   * service thread by placing this object on the service queue.
   ********************************************************************/
   public final class AutoCloseWorkItem extends CassetteWorkItem {

      /*****************************************************************
      * This AutoCloseWorkItem attribute contains a batch close request
      * for the batch that is to be automatically closed.  It is used by
      * the parent class to determine what framework locks must be
      * obtained by the thread before it is safe to call the doYourWork()
      * method.  It is also used by this object to obtain references
      * to the framework and cassette objects needed to perform
      * the close request.
      *****************************************************************/
      private BatchCloseRequest objRequest = null;

      /*****************************************************************
      * This constructor initializes this work item and adds it to the
      * frameworks service thread queue.
      *
      * @param     argRequest - a BatchCloseRequest for the batch
      * to be closed automatically.
      * @param     argCassette - the cassette that owns the batch.
      * @exception ETillAbortOperation - any failure results in this
      * exception.
      *****************************************************************/
      public AutoCloseWorkItem(
         BatchCloseRequest argRequest,
         Cassette          argCassette
      ) throws ETillAbortOperation {

         // Initialize the parent.  The parent will examine the
         // supplied request object to determine what locks must
         // be obtained before this work item can be performed.
         super(argRequest, argCassette);

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionEntry(
               TRACE_ID,
               " BatchAutoClose.BatchAutoClose()"
            );
         }

         // Save the batch close request for future reference.
         this.objRequest = argRequest;

         // Add this work item to the service queue.
         Supervisor.addItemToServiceQueue(this);

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionExit(
               TRACE_ID,
               " BatchAutoClose.BatchAutoClose()"
            );
         }
      }

      /*****************************************************************
      * This method is called by the framework service thread to perform
      * the work assigned to this work item.  Before calling this
      * method, the service thread will obtain all the required locks.
      * The list of required locks is determined by the Request object
      * (in this case a BatchCloseRequest object) that was specified
      * when this work item was constructed.  See the framework Javadoc
      * for the BatchCloseRequest object to see what locks are obtained.
      *
      * @exception ETillAbortOperation - any failure results in this
      * exception.
      *****************************************************************/
      public void doYourWork() throws ETillAbortOperation {
         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionEntry(
               TRACE_ID,
               " BatchAutoClose.doYourWork()"
            );
         }

         // Initialize variables to hold the result.
         short     varPrimaryRc       = 0;
         short     varSecondaryRc     = 0;
         Exception varExceptionCaught = null;

         // Call the cassette's service() method to close the
         // batch.
         try {
            APIResponse varResponse = new APIResponse(
               this.objRequest.getConnection()
            );
            this.getCassette().service(
               this.objRequest,
               varResponse
            );
            // No exception was thrown.  Extract the results from
            // the response.
            varPrimaryRc = varResponse.getPrimaryRC();
            varSecondaryRc = varResponse.getSecondaryRC();
         } catch (ETillAbortOperation varException) {
            // An ETillAbortOperation exception was thrown.  Extract
            // the results from the exception.
            varPrimaryRc = varException.getPrimaryRC();
            varSecondaryRc = varException.getSecondaryRC();
            varExceptionCaught = varException;
         }

         // Regardless of whether or not an exception was thrown,
         // examine the results and log them if an error occured.
         if (varPrimaryRc != PRC_OPERATION_SUCCESS) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardBatch Automatic BatchClose failed." +
                  "\n\tPRC = " + varPrimaryRc +
                  "\n\tSRC = " + varSecondaryRc
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_BATCH_AUTO_CLOSE_FAILURE,
               varExceptionCaught,
               this.objRequest.getBatch().getMerchantName(),
               this.objRequest.getBatchNumber(),
               this.objRequest.getAccountNumber(),
               Integer.toString(varPrimaryRc),
               Integer.toString(varSecondaryRc)
            );
            // Throw an abort operation for the service thread.
            throw new ETillAbortOperation(
               varPrimaryRc,
               varSecondaryRc
            );
         }

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionExit(
               TRACE_ID,
               " BatchAutoClose.doYourWork()"
            );
         }
      }
   }

   /********************************************************************
   * The TimeOfDayWorkItem object is used to automatically close an
   * open batch at a particular time of day (specified in minutes after
   * midnight).  When constructed, it automatically schedules itself
   * using the framework's timer queue.  When the scheduled time is
   * reached, the framework gives this object to the next available
   * service thread by placing this object on the service queue.
   ********************************************************************/
   public final class TimeOfDayWorkItem extends CassetteWorkItem {

      /*****************************************************************
      * This TimeOfDayWorkItem attribute contains a batch close request
      * for the batch that is to be automatically closed at a specified
      * time of day.  It is used in two ways.  It is used by the
      * parent class to determine what framework locks must be obtained
      * by the thread before it is safe to call the doYourWork()
      * method.  It is also used by this object to obtain references
      * to the framework and cassette objects needed to perform
      * the close request.
      *****************************************************************/
      private BatchCloseRequest objRequest = null;

      /*****************************************************************
      * This constructor
      * <UL>
      * <LI>initalizes a TimeOfDayWorkItem
      * <LI>constructs a new TimeableTransaction that references this
      * TimeOfDayWorkItem as well as the time when it is to be triggerd
      * <LI>adds the TimeableTransaction to the frameworks timer queue.
      * </UL>
      *
      * @param     argRequest - a BatchCloseRequest for the batch
      * to be closed automatically. (NOTE:  This batch close request
      * cannot be used to actually close the batch because it does
      * not contain locks for all the orders that are added to
      * the batch after it is constructed and before the batch close
      * is actually performed.)
      * @param     argCassette - the cassette that owns the batch.
      * @exception ETillAbortOperation - any failure results in this
      * exception.
      *****************************************************************/
      public TimeOfDayWorkItem(
         BatchCloseRequest argRequest,
         Cassette          argCassette
      ) throws ETillAbortOperation {

         // Initialize the parent.  The parent will examine the
         // supplied request object to determine what locks must
         // be obtained before this work item can be performed.
         super(argRequest, argCassette);

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionEntry(
               TRACE_ID,
               " TimeOfDayWorkItem.TimeOfDayWorkItem()"
            );
         }

         // Save the batch close request for future reference.
         this.objRequest = argRequest;
         // Obtain the framework batch.
         Batch varFrameworkBatch =
            this.objRequest.getBatch();
         // Obtain the cassette batch.
         LdbCardBatch varCassetteBatch =
            (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
         // Obtain the account used by the batch.
         LdbCardAccount varCassetteAccount =
            varCassetteBatch.getAccount();
         // Obtained the configured close time and add this work
         // item to the timer queue.
         this.addToTimerQueue(varCassetteAccount.getBatchCloseTime());

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionExit(
               TRACE_ID,
               " TimeOfDayWorkItem.TimeOfDayWorkItem()"
            );
         }
      }

      /*****************************************************************
      * This method schedules this work item by creating a new
      * TimeableTransaction that references this object as well as the
      * time when it is to be triggered and adding the
      * TimeableTransaction to the framework's timer queue.
      *
      * @param     argMinutesAfterMidnight - the minutes after midnight
      * when the automatic batch close is to be performed.
      *****************************************************************/
      private void addToTimerQueue(
         int argMinutesAfterMidnight
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionEntry(
               TRACE_ID,
               " TimeOfDayWorkItem.addToTimerQueue()"
            );
         }

         // Create a calandar object to be used as a date/time
         // calculator and initialize it with the current date/time.
         GregorianCalendar varDateTimeCalculator = new GregorianCalendar();
         // Get the current date/time in milliseconds for comparison.
         long varThisInstant = varDateTimeCalculator.getTime().getTime();
         // By definition, midnight is the first millisecond of the
         // new day.  Set today's midnight into the calculator.
         varDateTimeCalculator.set(
            GregorianCalendar.HOUR,
            0
         );
         varDateTimeCalculator.set(
            GregorianCalendar.MINUTE,
            0
         );
         varDateTimeCalculator.set(
            GregorianCalendar.SECOND,
            0
         );
         varDateTimeCalculator.set(
            GregorianCalendar.MILLISECOND,
            0
         );
         // Add the configured minutes after midnight to the value
         // in the calculator to compute the time when the automatic
         // batch close would be performed if it was to be performed
         // today.
         varDateTimeCalculator.add(
            GregorianCalendar.MINUTE,
            argMinutesAfterMidnight
         );
         // Convert the proposed time into milliseconds.
         long varBatchCloseTime = varDateTimeCalculator.getTime().getTime();
         // Check to see if the proposed time has already passed.
         if (varBatchCloseTime < varThisInstant) {
            // Since the proposed time has already passed, add 1 day to
            // the calculator to compute the correct date/time for the
            // automatic close to be performed tomorrow.
            varDateTimeCalculator.add(
               GregorianCalendar.DATE,
               1
            );
            // Convert the proposed time into milliseconds.
            varBatchCloseTime = varDateTimeCalculator.getTime().getTime();
         }
         // Use a new TimeableTransaction to schedule this work item
         // to be peformed at the proper time.
         Supervisor.addItemToTimerQueue(
            new TimeableTransaction(
               this,
               varBatchCloseTime
            )
         );


         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionExit(
               TRACE_ID,
               " TimeOfDayWorkItem.addToTimerQueue()"
            );
         }
      }

      /*****************************************************************
      * This method is called by the framework service thread to perform
      * the work assigned to this work item.  Before calling this
      * method, the service thread will obtain all the required locks.
      * The list of required locks is determined by the Request object
      * (in this case a BatchCloseRequest object) that was specified
      * when this work item was constructed.  See the framework Javadoc
      * for the BatchCloseRequest object to see what locks are obtained.
      *
      * @exception ETillAbortOperation - any failure results in this
      * exception.
      *****************************************************************/
      public void doYourWork() throws ETillAbortOperation {
         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionEntry(
               TRACE_ID,
               " TimeOfDayWorkItem.doYourWork()"
            );
         }

         // Obtain the framework batch.
         Batch varFrameworkBatch =
            this.objRequest.getBatch();
         // Obtain the cassette batch.
         LdbCardBatch varCassetteBatch =
            (LdbCardBatch)varFrameworkBatch.getCassetteBatch();
         // Obtain the cassette account.
         LdbCardAccount varCassetteAccount =
            varCassetteBatch.getAccount();

         // Make sure the batch is still a candidate for an automatic
         // batch close.  If the batch is closed to new transactions,
         // this automatic close request has lost a race and some
         // other request (perhaps the merchant explicitly closed the
         // batch, perhaps the batch was closed because it reached its
         // maximum size) has already closed the batch so this work item
         // can be safely abandoned.
         if (!varCassetteBatch.isBatchClosedToNewTransactions()) {

            // Make sure the merchant still wants the batch to be
            // closed automatically.  If the merchant modified the
            // configuration so that an automatic time-of-day close
            // is no longer configured, this work item can be safely
            // abandoned.
            if (varCassetteAccount.isBatchCloseTimeConfigured()) {

               // See if the merchant configured the account so that an
               // automatic time-of-day close is performed only when
               // the batch contains sufficent funds to justify
               // settlement.
               if (varCassetteAccount.isMinBatchTotalConfigured()) {

                  // Obtain the batch item lists and compute the
                  // value of the batch.  NOTE:  This code assumes
                  // that there is only one currency in the batch.
                  // If there is more than one currency in the batch
                  // some type of currency conversion must be done
                  // to compute a value that can be correctly
                  // compared with the configured minimum batch
                  // total amount.
                  Enumeration varItemList =
                     varFrameworkBatch.getTotals();
                  int varBatchValue = 0;
                  while (varItemList.hasMoreElements()) {
                     BatchTotalsItem varItem =
                        (BatchTotalsItem)varItemList.nextElement();
                     varBatchValue +=
                        varItem.getPaymentTotal() -
                        varItem.getCreditTotal();
                  }

                  // Test if the batch contains sufficient funds to
                  // justify settlement.
                  if (varBatchValue >= varCassetteAccount.getMinBatchTotal()) {

                     // There are sufficient funds to justify the
                     // settlement, so perform the automatic batch
                     // close.
                     this.scheduleAutoClose(varCassetteBatch);

                  } else {

                     // There are insufficient funds to justify the
                     // settlement, so reschedule this work item for
                     // tomorrow.
                     this.addToTimerQueue(
                        varCassetteAccount.getBatchCloseTime()
                     );
                  }

               } else {

                  // There is no minimum batch total configured so
                  // perform the automatic batch close.
                  this.scheduleAutoClose(varCassetteBatch);

               }

            }
         }

         if (Trace.isAnyoneTracing()) {
            Trace.traceFunctionExit(
               TRACE_ID,
               " TimeOfDayWorkItem.doYourWork()"
            );
         }

      }

      /*****************************************************************
      * This method is used to schedule an automatic batch close.  The
      * BatchClose request used to construct this cassette work item
      * guarantees that the write lock of the framework Account object
      * has been acquired before this workitem was given control.
      * This guarantees that no Deposit, DepositReversal, Refund, or
      * RefundReversal commands (in other words, the commands that
      * could change the contents of a batch) can race in at this
      * point.  That is why no synchronization is necessary to
      * schedule the automatic close.
      * <p>
      * Note: The automatic batch close cannot be performed on this
      * thread because this thread does not have all the required
      * locks.  Most likely, payments and credits have been added to
      * the batch since this object's BatchCloseRequest was constructed.
      * Therefore, this object's BatchCloseRequest does not contain
      * references to the order object locks for these additional
      * payments and credits.  A new BatchCloseRequest (one that
      * represents the current state of the batch) must be constructed
      * and sent to a service thread where the proper locks can be
      * obtained before performing the close.
      *
      * @param     argCassetteBatch - the batch to be closed.
      * @exception ETillAbortOperation - any failure results in this
      * exception.
      *****************************************************************/
      private void scheduleAutoClose(
         LdbCardBatch argCassetteBatch
      ) throws ETillAbortOperation {

         // Ask the batch to close itself using a service thread.  This
         // is required because a different set of locks (specifically
         // the order locks for all the orders with payments or credits
         // currently in the batch) must be obtained before the
         // batch can be closed.
         argCassetteBatch.scheduleAutoClose();
         // Commit the modified objects to the database.
         Supervisor.getThreadCommitPoint().commit();
      }

   }

}
